# -*- coding: utf-8 -*-

from blocks.simulink_basic import *
from utilities.simulink_utilities import *

# Structure for an Inport/Outport  block
class SimulinkInOutPort(SimulinkBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		self.create_ports(1, 1)
		# each inport and outport has a unique index for a subsystem (used to connect ports)
		self.index = int(get_property(self.xml, 'Port')) 

	def update_arities(self):
		SimulinkBlock.update_arities(self)
		
		# set by default the arity of an Inport block to 1
		if self.get_type() == "Inport" and self.get_inport_arity(0) == 0:
			self.set_inport_arity(0, 1)

	def get_index(self):
		return self.index


# Structure for an ActionPort block (from If Action Subsystem)
class SimulinkActionPort(SimulinkBlock):
	index = 0
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		self.create_ports(2, 1)
	
	def update_arities(self):
		flag = 0
		for i in range(len(self.outport)):
			if self.get_inport_arity(i+1) != self.get_outport_arity(i):
				flag = 1
				self.set_outport_arity(i, self.get_inport_arity(i+1))
		return flag

	def update_const_params(self, opt_const, opt_generic):
		p = "ve_" + str(SimulinkActionPort.index)
		self.set_params(p)
		SimulinkActionPort.index += 1 

	def get_def_params(self, opt_fun, opt_generic):
		params = self.get_params()
		if params:
			return "(%s::'a::simulink)" % params
		else:
			return params

	
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name(), self.get_params_list())
		
		invars  = [x.get_var_list() for x in self.inport]
		flat_invars = [j for i in invars for j in i]
		cond = flat_invars[0]
		expr1 = ",".join([i for i in flat_invars[1:]])
		expr2 = ",".join(["0" for i in flat_invars[1:]])
		if self.get_params():
			zero = "(0::'a)"
		else:
			zero = "0"
		expr = "If %s \<noteq> %s Then (%s) Else (%s)" % (cond, zero, expr1, expr2)
			
		
		return iFunc(flat_invars, expr, opt_fun)

	def split_block(self):
		n = len(self.inport)
		assert(n == len(self.outport) + 1)
		skip_blocks = []
		for i in range(1, n):
			skip_blocks.append(SimulinkActionPortSingle(self, i, [self.inport[0], self.inport[i]], [self.outport[i - 1]]))
		return skip_blocks


# Structure for ActionPort variation with one output
class SimulinkActionPortSingle(SimulinkActionPort):
	def __init__(self, actionport, i, inport, outport):
		SimulinkBlock.__init__(self)
		self.inport = inport
		self.outport = outport
		self.parent = actionport.parent
		self.name = actionport.get_name() + str(i)
		# splitted instances of the same actionport should have the same parameter(s)
		self.params = actionport.params


# Structure for a SwitchCase block 
class SimulinkSwitchCase(SimulinkBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)

		condition = get_property(self.xml, "CaseConditions")
		self.condition = parse_conj(condition)
		self.default_case = get_property(self.xml, "ShowDefaultCase") == "on"

	def get_var(self):
		return self.inport[0].get_var_name()

	def get_cases(self):
		var = self.get_var()
		cases = [mk_or(var, "=", e) for e in self.condition]
		if self.default_case:
			all_exp = [e for x in self.condition for e in x]
			cases.append(mk_and(var, "\<noteq>", all_exp))
		return cases
		
	def isabelle(self, opt_fun, opt_generic, flag):
		expr = ", ".join(self.get_cases())
		return iFunc([self.get_var()], expr, opt_fun)

	def split_block(self):
		cases = self.get_cases()
		return [SimulinkSwitchCaseSingle(self, cases[i], self.outport[i], i+1) for i in range(len(cases))]


# Structure for SwitchCase variant with one output
class SimulinkSwitchCaseSingle(SimulinkBlock):
	def __init__(self, switch, case, outport, index):
		SimulinkBlock.__init__(self)
		self.inport = switch.inport
		self.outport = [outport]
		self.case = case
		self.var = switch.get_var()
		self.name = switch.get_name() + str(index)
		self.parent = switch.parent
				
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())
			return self.get_qualified_name()

		#todo: check if this is good enough. remove function mk_update from simulink_utilities.py
		return iFunc([self.var], self.case, opt_fun)
		return mk_update(self.var, "(" + self.case +")")

	
# TODO: Simulink Enable block


	
