# -*- coding: utf-8 -*-

from blocks.simulink_basic import *
from blocks.simulink_continuous import *
from blocks.simulink_discontinuous import *
from blocks.simulink_discrete import *
from blocks.simulink_logic import *
from blocks.simulink_math import *
from blocks.simulink_ports import *
from blocks.simulink_sigrouting import *
from blocks.simulink_sinks import *
from blocks.simulink_sources import *
from blocks.simulink_lookup import *
from blocks.simulink_verification import *

from utilities.simulink_parser import *
from utilities.simulink_utilities import *
from strategies.simulink_composition import *
from pprint import pprint
import utilities.simulink_globals



# Structure for a wire split
class Split(SimulinkBlock):
	def __init__(self, index, size_out, parent = None):
		
		size_in = 1
		SimulinkBlock.__init__(self, xml = None, size_in = size_in, size_out = size_out, parent = parent)
		self.sid = "S%d" % index
		self.name = "Split%d" % index
		self.type = "Split"
	
	
	def update_arities(self):
		flag = 0
		a = self.get_inport_arity(0)
		for p in self.outport:
			if p.get_arity() != a:
				p.set_arity(a)
				flag = 1
		return flag

	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())
			
		n = len(self.inport)
		m = len(self.outport)
		if m % n > 0:
			raise Exception('\nError:\n  For a wire split block, the number of out ports must be a multiple of the in ports.\n')
		return iSwitchSplit([self.inport], [self.inport * (m / n)], opt_fun)
		
	def split_block(self):
		# change ports with multiplicities into single ports
		n = len(self.inport)
		m = len(self.outport)
		if m % n > 0:
			raise Exception('\nError:\n  For a wire split block, the number of out ports must be a multiple of the in ports.\n')
			
		k = m / n

		skip_blocks = []
		x = 0
		for j in range(k):
			for i in range(n):
				skip_blocks.append(Skip(self.inport[i], self.outport[x + i]))
			x = x + n
		return skip_blocks



# Defines a class corresponding to a Simulink System
# Model and SubSystem are extensions of System
class SimulinkSystem(SimulinkBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, 0, 0, parent)

		print 'Initializing component: ', self.get_qualified_name() #self.sid
		
		# here xml holds a Model or a SubSystem, and both have a System component
		curComponent = xml.find('System')
		
		if curComponent == None:
			print "The component must contain a System!"
			exit(0)
		
		new_top_blocks = {}
		#self.fblocks = []	
		
		# 1. instantiate blocks contained by the current subsystem
		# in case of hierarchical structure, recursive application of the instantiation
		
		top_blocks = curComponent.findall('Block')
		for tb in top_blocks:
			block = create_simulink_block(tb, self)
			new_top_blocks[block.get_id()] = block
			#if block.get_type() == 'Fcn':
			#	self.fblocks.append(block)

		new_lines = []
		

		# 2. build wires from xml structure 
		top_lines = curComponent.findall("./Line")
		
		index_src = 0
		index_dest = 0
		
		while top_lines != []:
			line = top_lines.pop(0)
			src = get_property(line, 'Src')
			if src:
				src = src.split("#")
				x = src[1].split(":")
				src = (src[0], x[0], int(x[1]) - 1)
			else:
				src = ("Src", "out", index_src)
				index_src += 1
				try:
					dst = get_property(line, 'Dst').split("#")
					print "\nWarning:\n\tThe input wire for block "+ utilities.simulink_globals.sid2element[dst[0]].attrib['Name'] +" of type "+ utilities.simulink_globals.sid2element[dst[0]].attrib['BlockType'] +" is not connected.\n\tIgnoring this wire in the translation."
				except:
					print "\nWarning:\n\tAn input unconnected wire to a Split has been detected in the model.\n\tIgnoring this wire in the translation."
				continue
			
			block_type = 'Unconnected' if src[0] == 'Src' else ('Split-' + src[0] if src[0].startswith('S') else utilities.simulink_globals.sid2element[src[0]].attrib['BlockType'])
			
			if not (get_property(line, 'Dst') is None) :
				dst = get_property(line, 'Dst').split("#")
				x = dst[1].split(":")
				#dst = (dst[0], x[0], int(x[1]) - 1)
				if len(x) > 1:
					dst = (dst[0], x[0], int(x[1]) - 1)
				else:
					dst = (dst[0], "in", x[0])
				new_lines.append((src, dst))
			else:
				branches = line.findall('./Branch')
				if branches:
					split_block = Split(len(new_top_blocks) + 1, len(branches), self)
					
					#split_block.set_qualified_name(self.get_qualified_name())
					
					set_property(line, 'Dst', '%s#in:1' % split_block.get_id())
					dst = (split_block.get_id(), "in", 0)
					new_lines.append((src, dst))
					
					for i, branch in enumerate(line.findall('./Branch')):
						set_property(branch, 'Src', '%s#out:%d' % (split_block.get_id(), i + 1))
						top_lines.append(branch)

					new_top_blocks[split_block.get_id()] = split_block
				else:
					# print "Warning: an unconnected wire has been detected in the model. Proceeding without it."
					if src[0].startswith('S'):
						print "\nWarning:\n\tAn output unconnected wire of a Split has been detected in the model.\n\tIgnoring this wire in the translation.\n"
					else:
						print "\nWarning:\n\tThe output wire of block " + utilities.simulink_globals.sid2element[src[0]].attrib['Name'] + " of type " + utilities.simulink_globals.sid2element[src[0]].attrib['BlockType']+" is not connected.\n\tIgnoring this wire in the translation."
					dst = ("Dst", "in", index_dest)
					index_dest += 1
					# the following line builds wires that do not have a destination. 
					# these wires are not concerned by the translation, therefore it is commented out
					#new_lines.append((src, dst))


		# build lines for Goto and From blocks
		for i in new_top_blocks.keys():
			src = new_top_blocks[i]
			if src.get_type() == "Goto":
				# get corresponding From blocks
				dests = []
				for j in new_top_blocks.keys():
					if new_top_blocks[j].get_type() == 'From' and new_top_blocks[j].label == src.label:
						dests.append(j)
				# create output ports for each From block
				src.add_ports(len(dests)-1)
				# link ports
				for j in range(len(dests)):
					tsrc = (i, "out", j)
					tdest = (dests[j], "in", 0)
					new_lines.append((tsrc, tdest))

						
		# 3. instantiate wires: add missing ports to blocks and replace input ports
		# with corresponding connected output ports.
		
		#print "handling wires"

		for line in new_lines:
			#print line
			(bid, io, index) = line[0]
			p = Port(bid, index, io, 0)
			## print "(%s, %s, %s)" %(bid, io, index)
			if io == "out":
				new_top_blocks[bid].set_out(index, p)
			else:
				print "unknown port out"
				
			#handle case where index is ifaction
			(bid, io, index) = line[1]
			#print "(%s, %s, %s)" %(bid, io, index)
			if io == "in":
				#replace the input port with corresponding connected output port
				#print bid
				new_top_blocks[bid].set_in(index, p)
			else:
				#print (bid, io, index)
				print "unknown port in"
		
		for kb in new_top_blocks.keys():
			block = new_top_blocks[kb]
			block.add_state_ports()
			
			# set variable names for ports: a, b, ..., aa, ba, ca, ...
			block.set_variables(utilities.simulink_globals.variable_names)


		#todo: think if we need it as a dictionarry
		keys = new_top_blocks.keys()
		self.blocks = [new_top_blocks[bk] for bk in keys]
		#print self.fblocks
		
	def split_ports(self):
		SimulinkBlock.split_ports(self)
		for b in self.blocks:
			b.split_ports()
			
	def split_block(self):
		blocks = self.blocks
		self.blocks = []
		for b in blocks:
			self.blocks.extend(b.split_block())
	
		return [self]

	def get_fblocks(self,res):	
		#if len(self.fblocks) != 0:
		#	res += [x for x in self.fblocks]
		for b in self.blocks:
			if isinstance(b, SimulinkFunction):
				res.append(b)
			if isinstance(b, SimulinkSubsystem):
				x = b.get_fblocks(res)
				#print x
				#res.append(x)
	
	def get_bblocks(self, res):
		for b in self.blocks:
			if isinstance(b, SimulinkSubsystem):
				b.get_bblocks(res)
			else:
				res.append(b)

	def get_sample_dt(self, vals):
		for b in self.blocks:
			if isinstance(b, SimulinkSubsystem):
				b.get_sample_dt(vals)
			else:
				if b.get_type() in ["Delay", "UnitDelay"]:
					vals[b] = b.sample_dt

	def update_const_params(self, opt_const, opt_generic):
		for b in self.blocks:
			b.update_const_params(opt_const, opt_generic)

	def update_generic_params(self):
		for b in self.blocks:
			b.update_generic_params()

	def update_const_vals(self, cvals):
		for b in self.blocks:
			b.update_const_vals(cvals)

	def set_init(self):
		for b in self.blocks:
			b.set_init()
			
	def get_nr_blocks(self):
		s = 0
		for b in self.blocks:
			s = s + b.get_nr_blocks()
		return s


# Structure for a Subsystem block
class SimulinkSubsystem(SimulinkSystem):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkSystem.__init__(self, xml, size_in, size_out, parent)
		self.definition_blocks = []

	def set_in(self, index, value):
		if index in ["ifaction", "enable"]:
			for b in self.blocks:
				if isinstance(b, SimulinkActionPort):
					b.set_in(0, value)
			return
			
		#update structure port
		n = len(self.inport)
		if index >= n:
			self.inport += [Port(self.sid, i, "in", 0) for i in range(n, index + 1)]
		self.inport[index] = value   
		
		#update inner component port
		for b in self.blocks:
			if isinstance(b, SimulinkInOutPort) and (b.get_type() == "Inport") and ((index + 1) == b.get_index()):
				b.set_in(0, value)
							
	def set_out(self, index, value):
		#update subsystem port
		n = len(self.outport)
		if index >= n:
			self.outport += [Port(self.sid, i, "out", 0) for i in range(n, index + 1)]
		self.outport[index] = value   
		
		#update inner component port
		for b in self.blocks:
			if isinstance(b, SimulinkInOutPort) and (b.get_type() == "Outport") and ((index + 1) == b.get_index()):
				b.set_out(0, value)

	def update_arities(self):
		flag = 0
		for b in self.blocks:
			if b.update_arities() == 1:
				flag = 1
		return flag

	def flatten_block_list(self, res):
		for b in self.blocks:
			if isinstance(b, SimulinkSubsystem):
				b.flatten_block_list(res)
			else:
				res.append(b)

			
	def compose(self, opt_comp, opt_fun):
		new_blocks = []
		for b in self.blocks:
			if isinstance(b, SimulinkSubsystem):
				new_blocks.append(b.compose(opt_comp, opt_fun))
			else:
				new_blocks.append(b)
		#self.blocks = new_blocks

		if opt_comp == "-fp":
			aux = sort_blocks2(new_blocks)
			self.formula = MonolithicComposition(self.definition_blocks, "MC" + self.sid, self.name, aux, opt_fun)
			self.formula.set_name(self.get_qualified_name())
			self.definition_blocks.append(self.formula)
			#print self.formula
		elif opt_comp == "-ic":
			aux = sort_blocks2(new_blocks)
			self.formula = CSerialComposition('CSC' + str(self.sid), self.name, aux)
			self.formula.set_name(self.get_qualified_name())
		else:
			aux = sort_blocks2(new_blocks)[::-1]
			self.formula = ISerialComposition(self.definition_blocks, 'ICS' + str(self.sid), self.name,  aux)
			self.formula.set_name(self.get_qualified_name())
			self.definition_blocks.append(self.formula)
		
		return self.formula
				
	def __repr__(self):
		res = '{Block} %s, %s\n' % (self.get_id(), self.get_type())
		return res
		
		res = '~~\n{Block} %s, %s, %s, %s\nContained blocks:\n' % (self.get_id(), self.get_type(), self.inport, self.outport) #, self.isabelle(False, True))
		for b in self.blocks:
			res = res + repr(b)
		res = res + "~~\n"
		return res
		
	
# Structure for a Model
class SimulinkModel(SimulinkSystem):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkSystem.__init__(self, xml, size_in, size_out, parent)
		self.definition_blocks = []
		self.set_name("Model")

	def update_arities(self):
		flag = 1
		
		max_iter = self.get_nr_blocks()
		counter = 1

		for b in self.blocks:
			b.update_arities()

		while flag == 1 and counter <= max_iter:
			flag = 0
			counter += 1
			#print 'update cycle'
			for b in self.blocks:
				if b.update_arities() == 1:
					flag = 1
					
		if counter > max_iter:
			raise Exception('\nError:\n  The wires arities could not be computed on this model.\n  Please check the model for infinite feedback loops.\n')
		
	def get_type(self):
		return "Model"
	
	def flatten_block_list(self, res):
		for b in self.blocks:
			if isinstance(b, SimulinkSubsystem):
				b.flatten_block_list(res)
			else:
				res.append(b)
				
				
	def compose(self, opt_comp, opt_flat, opt_fun):
		new_blocks = []
		if opt_flat:
			self.flatten_block_list(new_blocks)
		else:	
			for b in self.blocks:
				if isinstance(b, SimulinkSubsystem):
					new_blocks.append(b.compose(opt_comp, opt_fun))
				else:
					new_blocks.append(b)
		if opt_comp == "-fp":
			aux = sort_blocks2(new_blocks)		
			self.formula = MonolithicComposition(self.definition_blocks, 'MC' + str(self.sid), 'FP_Model', aux, opt_fun)
			#self.formula = MonolithicComposition(self.definition_blocks, 'MC' + str(self.sid), 'FP_Model', new_blocks, opt_fun)
		elif opt_comp == "-ic":
			aux = sort_blocks2(new_blocks)
			self.formula = CSerialComposition('CSC' + str(self.sid), 'IC_Model', aux)
		else:
			aux = sort_blocks2(new_blocks)[::-1]
			self.formula = ISerialComposition(self.definition_blocks, 'ICS' + str(self.sid), 'ICI_Model', aux)
			self.formula.set_name("ICI_Model")
			

# Structure for Enabled Subsystems			
class SimulinkSubsystemIf(SimulinkSubsystem):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkSystem.__init__(self, xml, size_in, size_out, parent)
		
		#print "initializing if subsystem"
		#make connection for ActionPort
		outports = []
		for b in self.blocks:
			if isinstance(b, SimulinkActionPort):
				ap_block = b
			if b.get_type() == "Outport":
				#print b
				outports.append(b)
				
		for b in outports:
			ap_block.set_in(b.get_index(), b.outport[0])
		
		self.definition_blocks = []
			
	def set_in(self, index, value):
		if index in ["ifaction", "enable"]:
			for b in self.blocks:
				if isinstance(b, SimulinkActionPort):
					b.set_in(0, value)
		else:
			SimulinkSubsystem.set_in(self, index, value)
			
	def set_out(self, index, value):
		for b in self.blocks:
			if isinstance(b, SimulinkActionPort):
				ap_block = b
		ap_block.set_out(index, value)	
		


		
SimulinkMapping = {
	"Scope"                  : SimulinkScope, 
	"Inport"                 : SimulinkInOutPort, 
	"Outport"                : SimulinkInOutPort,
	"Constant"               : SimulinkConstant,
	"Logic"                  : SimulinkLogic,
	"Fcn"                    : SimulinkFunction,
	"Mux"                    : SimulinkMux,
	"Sum"                    : SimulinkMath,
	"Product"                : SimulinkMath,
	"UnaryMinus"             : SimulinkMath,
	"Math"                   : SimulinkMathFunction,
	"Gain"                   : SimulinkGain,
	"Saturate"               : SimulinkSaturate,
	"Switch"                 : SimulinkSwitch,
	"UnitDelay"              : SimulinkDelay,
	"Delay"                  : SimulinkDelay,
	"Relay"                  : SimulinkRelay,
	"Integrator"             : SimulinkIntegrator,
	"Sqrt" 		             : SimulinkBlock,
	"SubSystem"              : SimulinkSubsystem,
	#"SubSystemIf"           : SimulinkSubsystemIf,
	#"SubSystemEnable"       : SimulinkSubsystemEnable,
	"DataTypeConversion"     : SimulinkDataTypeConversion,
	"Selector"               : SimulinkSelector,
	"Step"                   : SimulinkStep,
	"TransferFcn"            : SimulinkTransferFcn,
	"SwitchCase"             : SimulinkSwitchCase,
	"MultiPortSwitch"        : SimulinkMultiPortSwitch,
	"Demux"                  : SimulinkDemux, 
	"RelationalOperator"     : SimulinkRelationalOperator,
	"DiscretePulseGenerator" : SimulinkDiscretePulseGenerator,
	"ActionPort"             : SimulinkActionPort,
	"Terminator"             : SimulinkTerminator,
	"MinMax"                 : SimulinkMinMax,
	"EnablePort"             : SimulinkActionPort,
	"Goto"                   : SimulinkGoto,
	"From"                   : SimulinkFrom,
	"Sin"                    : SimulinkSinWave,
	"Stop"                   : SimulinkScope,
	"Lookup_n-D"             : SimulinkLookup1D,
	"SignalGenerator"        : SimulinkSignalGenerator,
	"Assertion"              : SimulinkAssertion,
	"CheckStaticRange"       : SimulinkCheckStaticRange,
	"CheckStaticGap"         : SimulinkCheckStaticGap,
	"CheckStaticLowerBound"  : SimulinkCheckStaticLowerBound,
	"CheckStaticUpperBound"  : SimulinkCheckStaticUpperBound,
	"CheckDynamicRange"      : SimulinkCheckDynamicRange,
	"CheckDynamicGap"        : SimulinkCheckDynamicGap,
	"CheckDynamicLowerBound" : SimulinkCheckDynamicLowerBound,
	"CheckDynamicUpperBound" : SimulinkCheckDynamicUpperBound
	}


def create_simulink_block(xml, parent):
	block_type = xml.attrib['BlockType']
	ports = get_property(xml, "Ports")
	is_active = False
	if ports:
		is_active = (len(eval(ports)) > 2)
	if block_type in SimulinkMapping.keys():
		if block_type == "MultiPortSwitch":
			return SimulinkMultiPortSwitch(xml, None, None, None, parent)
		if block_type == "SubSystem" and is_active:
			return SimulinkSubsystemIf(xml, None, None, parent)
		return SimulinkMapping[block_type](xml, None, None, parent)
	if block_type == "Reference":
		n = xml.attrib['Name'].replace(' ', '').replace('\n', '')
		if n in SimulinkMapping.keys():
			return SimulinkMapping[n](xml, None, None, parent)
		ref_block_type = get_property(xml,'SourceType').replace('\n', '').strip()
		block_library = get_property(xml,'SourceBlock').replace('\n', '').strip()[:-len(ref_block_type)]
		raise Exception("\nError:\n  The block "+ n +" of type " + ref_block_type + " from library " + block_library +" is not supported.\n  Please check the list of supported block types (option -blocks).\n") #("Unknown Simulink block: " + block_type + " " + n) # "of type " + block_type
	n = xml.attrib['Name'].replace(' ', '').replace('\n', '')
	raise Exception("\nError:\n  The block "+ n +" of type " + block_type +" is not supported.\n  Please check the list of supported block types (option -blocks).\n") #Exception("Unknown Simulink block: " + block_type)

