#!/usr/bin/python
# -*- coding: utf-8 -*-


import xml.etree.cElementTree as ET

import zipfile as ZF
from pprint import pprint
import itertools
import re
import sys
import os
import httplib
import json
import time
from copy import deepcopy
import traceback

import utilities.simulink_globals 
from blocks.simulink_system import *
from strategies.simulink_composition import *
from strategies.simulink_no_feedback import *
from utilities.simulink_fwrite import *
from utilities.simulink_parser import *
from utilities.simulink_options import *

def init_xml_file(filename):
	# Extract xml content
	slxFile = ZF.ZipFile(filename)
	slxFile.extract('simulink/blockdiagram.xml', 'extracted')
	xmlData = ET.parse('extracted/simulink/blockdiagram.xml')
	return xmlData.getroot()

def read_constants(root):
	# Read constants defined in an annotation textbox from the xml content
	consts = {}
	for annElem in root.findall(".//Annotation/P[@Name='ClickFcn']"):
		annLines = [PP_annotation_line.parseString(line) for line in annElem.text.split('\n')]
		for line in annLines:
			if len(line) == 3:
				consts[line[0]] = (line[2].replace("true","True").replace("false","False"))
	return consts

def read_model(root):
	# Read model from the xml content
	model = root.find('Model')

	blockDefaultParams = model.findall('BlockParameterDefaults')[0].findall('Block')

	for b in blockDefaultParams:
		bprop = {}
		for p in b.findall('./P'):
			bprop[p.attrib['Name']] = p.text
		utilities.simulink_globals.block_parameter_defaults[b.attrib['BlockType']] = bprop
	
	for e in root.iter():
		for c in e:
			utilities.simulink_globals.child2parent[c] = e
		if 'SID' in e.attrib:
			utilities.simulink_globals.sid2element[e.attrib['SID']] = e
	return model

def init_system(filename, opt_consts, opt_generic):
	# Build the internal representation of the Simulink diagram
	
	# Get xml content
	root = init_xml_file(filename)
	
	# Extract relevant information
	model =  read_model(root)
	consts = read_constants(root)
	# update global variables name list with those already in use (from the constants defined in the annotation textbox)
	utilities.simulink_globals.variable_names.update_excluded_names([k for k in consts.keys()])

	# Instantiate the model
	res = SimulinkModel(model)

	# Instatiate the arrities of each wires (number of values transported by each wire)
	res.update_arities()
	
	# Split ports having arity more than one into multiple ports of arity one
	res.split_ports()
	
	# Set the initial value for blocks with state variables
	res.set_init()
	
	# Generate vc_x parameters (used with the -consts) for Constant blocks as a solution to the "Additional types" error (i.e., trapped variables)
	if opt_consts:
		res.update_const_params(opt_consts, opt_generic)
	elif opt_generic:
		res.update_const_params(opt_consts, opt_generic)
	else:
		pass
		
	# Update the values of Constant blocks that are defined with a variable
	res.update_const_vals(consts)
	'''
	if not opt_generic:
		res.update_generic_params()
	'''
	
	return (res, consts)
	
	
def compute_cpt(res, opt_comp, opt_flat, opt_fun):
	# Compute the CPT for the model and its subsystems recursively based on the selected translation option 

	if opt_comp == "-nfb":
		# Split blocks having more than one output into several blocks with one output each
		res.split_block()
		
		# Apply translation strategy
		A = NoFeedbackComposition(res)
		res = ProdSingle(A)
		
	# Maybe update function with the following structure	
	#elif opt_comp == "-mc":
	#	resa = MonolithicComposition(res, ...)
	#elif opt_comp == "-":
	#	resa = CSerialComposition(res, ...)
	
	else:
		# Apply translation strategy
		res.compose(opt_comp, opt_flat, opt_fun)
		
	# Return the new Simulink diagram with the corresponding CPTs
	return res

def write_cpt(oFilename, res, consts, opt_comp, opt_flat, opt_fun, opt_consts, opt_generic, opt_type, opt_iter, opt_sim):
	# Generate the Isabelle theory
	
	# Obtain required paths for Isabelle "imports" command
	currentFolder =  os.path.dirname(rnpath(__file__))
	
	# Write the theory header (title + imports command + lemmas identifiers used in simplification)
	write_theory_title(oFilename, currentFolder, opt_comp, opt_sim, opt_iter)
	write_theory_header(oFilename)
	
	# Write the definition of constants declared in the annotation textbox
	write_constants(consts, oFilename)
	
	# Obtain the atomic blocks from the model and write their definition
	bblocks = []
	res.get_bblocks(bblocks)
	bblocks.sort(key=SimulinkBlock.get_qualified_name)
	write_atomic_blocks(bblocks, opt_fun, opt_generic, oFilename)
	
	# Write additional Isabelle code used for simplification
	write_simp_theorems(oFilename, opt_fun, opt_iter)

	# Write recursively the CPTs
	write_isabelle(res,  opt_comp, opt_flat, opt_fun, opt_consts, opt_generic, opt_type, opt_iter, opt_sim, oFilename)

	# Write additional Isabelle code 
	write_theory_inter_body(oFilename)

	write_theory_footer(oFilename, res, opt_type, opt_sim)

	if opt_sim:
		write_sim_step_file(res, oFilename)

	# If -iter is set, write the iteration of the model's CPT (we obtain in Isabelle the property transformer)
	if opt_iter:
		write_formula_iter(res, oFilename)
		if opt_type:
			write_formula_iter_type(res, opt_type, oFilename)
	
	# Write the theory footer
	write_theory_end(oFilename)
	
	# If -sim is set, write the Python simulation script
	if opt_sim:
		write_sim_file(res, oFilename, currentFolder, consts)
	

# Main method
def run():

	# Obtain translation options and check correct usage 	
		
	argv = sys.argv[1:]
	
	opts = Options(argv)
	
	(OPT_COMP, OPT_FLAT, OPT_CONSTS, OPT_GENERIC, OPT_TYPE, OPT_ITER, OPT_EQUIV, OPT_SIM) = ("-"+opts.comp, opts.flat, opts.const, opts.generic, opts.type, opts.iter, opts.equiv, opts.sim)
	OPT_FUN = True

	INPUT_FILE = opts.input_file
	OPT_OUTPUT_FILE = opts.output_file
	
	opts.print_options(OPT_SIM)

	# Instantiate the global variables used in the translation
	utilities.simulink_globals.init()
	
	
	# Obtain the internal representation of the Simulink diagram (i.e., blocks, constants declared in a textbox, generic vc_x parameters used with the -consts option)
	(res, consts) = init_system(INPUT_FILE, OPT_CONSTS, OPT_GENERIC)
	

	if not OPT_EQUIV:
		
		# Compute the CPT for the model and each subsystem recursively depending of the OPT_COMP translation option
		cres = compute_cpt(res, OPT_COMP, OPT_FLAT, OPT_FUN)
		
		# Write the Isabelle theory
		write_cpt(OPT_OUTPUT_FILE, cres, consts, OPT_COMP, OPT_FLAT, OPT_FUN, OPT_CONSTS, OPT_GENERIC, OPT_TYPE, OPT_ITER, OPT_SIM)

	else:
		print
		
		# For each translation method
		
		aux = ['fp','ic','ici','nfb']
		cpts = []
		for opt_comp in aux:
			ires = deepcopy(res)
			
			# Compute the CPT for the model and each subsystem with
			cres = compute_cpt(ires, "-" + opt_comp, OPT_FLAT, OPT_FUN) 
			cres.formula.set_name(cres.formula.get_name()+"_"+opt_comp)
			cpts.append(cres)
			
			# Write the Isabelle theory
			OPT_OUTPUT_FILE_TEST = OPT_OUTPUT_FILE.replace(".thy","_"+ opt_comp+".thy")	
			write_cpt(OPT_OUTPUT_FILE_TEST, cres, consts, "-"+opt_comp, OPT_FLAT, OPT_FUN, OPT_CONSTS, OPT_GENERIC, OPT_TYPE, OPT_ITER, OPT_SIM)

			print "%s translation strategy completed" % opt_comp

		# Write Isabelle theory with equivalence lemmas		

		write_equiv_theory(OPT_OUTPUT_FILE, cpts)

try:
	run()
	print '\nTranslation successful!\n'
except BaseException as e:
	#print e
	#traceback.print_exc()
	exit(1)

	
exit(0)
