(* Theory to demonstrate the Refinement Calculus of Reactive Systems (RCRS). *)
(* Part of artifact for paper "The Refinement Calculus of Reactive Systems Toolset",
   by Dragomir, Preoteasa, Tripakis, to appear in TACAS 2018. *)
(* The Sections and Appendix below refer to Sections in the above paper. *)
(* An arXiv version of the paper can be found at: https://arxiv.org/abs/1710.08195 *)

theory RCRS_Demo imports "./RCRS/Isabelle/Simulink/SimplifyRCRS"
  "./RCRS/Isabelle/RefinementReactive/Temporal"

begin
  declare [[eta_contract=false]]
  
  named_theorems basic_simps
  lemmas [basic_simps] = simulink_simps

(* Definitions of RCRS components in Section 2: Modeling Systems in RCRS *)
  definition [basic_simps] : "MyId = [: x \<leadsto> y . y = x :]"
  definition [basic_simps] : "Add = [: (x, y) \<leadsto> z . z = x + y :]"
  definition [basic_simps] : "MyConstant c = [: x::unit \<leadsto> y . y = c :]"
  definition [basic_simps] : "UnitDelay = [: (x,s) \<leadsto> (y,s') . y = s \<and> s' = x :]"
  definition [basic_simps] : "SqrRoot = {. x. x\<ge>0 .} o [- x \<leadsto> sqrt x -]"
  definition [basic_simps] : "NonDetSqrt = {. x . x\<ge>0 .} o [: x \<leadsto> y. y\<ge>0 :]"
  definition [basic_simps] : "ReceptiveSqrt = [: x \<leadsto> y . x \<ge> 0 \<longrightarrow> y = sqrt x:]"

  definition "prop_x = (\<lambda> x . x (0::nat))"
  definition "prop_y = (\<lambda> x y . y (0::nat))"
  definition "A1 = {. (\<box> \<diamondsuit> prop_x) .} o [: \<box> \<diamondsuit> prop_y :]"

(* Two alternative definitions of the Integrator component in 
Section 3: The Implementation of RCRS in Isabelle *)
  definition "Integrator1 dt = [- (x,s) \<leadsto> (s, s+x*dt) -]"
  definition "Integrator2 dt = [: (x,s) \<leadsto> (y,s'). y=s \<and> s'=s+x*dt :]"


(* The rest follows the demonstration described in the paper in 
Appendix A.1: Basic Reasoning in RCRS *)

  definition [basic_simps] : "Cst1 = [- u \<leadsto> 1 -]"
  
  simplify_RCRS "Syst1 = Cst1 o SqrRoot"
    "u"
    "y"
    
  thm Syst1_simp
    
  definition [basic_simps] : "Cst2 = [- u \<leadsto> -1 -]"
  
  simplify_RCRS "Syst2 = Cst2 o SqrRoot"
    "u"
    "y"
    
  thm Syst2_simp
    
  lemmas [basic_simps] = simulink_simps comp_rel_simps
  
  definition [basic_simps] : "true = [: u \<leadsto> y::real. True :]"
  
  simplify_RCRS "Syst3 = true o SqrRoot"
    "u"
    "y"
    
  thm Syst3_simp
    
  lemma aux1: "(\<forall>y::real. 0 \<le>y) = False"
    (*sledgehammer *)
    using le_minus_one_simps(3) by blast
  
  simplify_RCRS "Syst4 = true o SqrRoot"
    "u"
    "y"
    use(aux1)
  
  thm Syst4_simp
    
  definition [basic_simps] : "A = [: x \<leadsto> y. y\<ge>x+1 :]"
  
  simplify_RCRS "Syst5 = A o SqrRoot"
    "x"
    "y"
    
  thm Syst5_simp
    
  lemma aux2: "(\<forall>y::real\<ge>x+1. 0\<le>y) = (x \<ge>-1)"
    by auto
    
  simplify_RCRS "Syst6 = A o SqrRoot"  
    "x"
    "y"
    use(aux2)
    
  thm Syst6_simp

  lemma power_mono: "0\<le>(x::real) \<Longrightarrow> x \<le> y \<Longrightarrow> x^2 \<le> y^2"
    by (rule Power.linordered_semidom_class.power_mono)
      
  lemma aux3:  "(\<lambda> x z. x \<ge> -1 \<and> (\<exists>y. y \<ge> (x + 1) \<and> z = sqrt y)) = (\<lambda> x z. (x \<ge>-1 \<and> z \<ge> sqrt (x+1)))"
    apply (simp add: fun_eq_iff)
    apply safe
    apply simp_all
    apply (rule_tac x="xa^2" in exI)
    apply safe
    apply simp_all
    apply (cut_tac x="sqrt (x + 1)" and y ="xa" in power_mono)
    apply simp_all
    apply (subgoal_tac "xa\<ge>0")
    apply (simp)
    by (metis add.commute order_trans real_0_le_add_iff real_sqrt_ge_0_iff)
      
  lemma aux4: "{. x. x \<ge> -1 .} o [: x \<leadsto> z. \<exists>y\<ge>x + 1. z = sqrt y :] = {. x. x \<ge> -1 .} o [: x \<leadsto> z.  \<exists>y. x \<ge>-1 \<and> y\<ge>x + 1 \<and> z = sqrt y :]"      
    by (auto simp add: fun_eq_iff assert_def demonic_def)
  
  lemma aux5: "{. x. x \<ge> -1 .} o [: x \<leadsto> z. z \<ge> sqrt (x+1) :] = {. x. x \<ge> -1 .} o [: x \<leadsto> z. x\<ge>-1 \<and> z \<ge> sqrt (x+1) :]"
    by (auto simp add: fun_eq_iff assert_def demonic_def)
      
  lemma "Syst6 = {. x. x \<ge> -1 .} o [: x \<leadsto> z. z \<ge> sqrt (x+1) :]"
    apply (simp add: Syst6_simp)
    apply (subgoal_tac "{. x. x \<ge> -1 .} o [: x \<leadsto> z. \<exists>y\<ge>x + 1. z = sqrt y :] = {. x. x \<ge> -1 .} o [: x \<leadsto> z.  \<exists>y. x \<ge>-1 \<and> y\<ge>x + 1 \<and> z = sqrt y :]")
    apply (simp_all add: aux4)
    apply (subgoal_tac "{. x. x \<ge> -1 .} o [: x \<leadsto> z. z \<ge> sqrt (x+1) :] = {. x. x \<ge> -1 .} o [: x \<leadsto> z. x\<ge>-1 \<and> z \<ge> sqrt (x+1) :]")
    apply (simp_all add: aux5)
    by (simp add: fun_eq_iff aux3)


  lemmas [basic_simps] = comp_rel_simps basic_block_rel_simps update_def refinement_simps
  
  lemma "NonDetSqrt \<le> SqrRoot"
    by (auto simp add: basic_simps)
        
  lemma "SqrRoot \<le> ReceptiveSqrt" 
    by (simp add: basic_simps)   

end